import importlib


def requires(module_name: str):
    try:
        importlib.import_module(module_name)
    except ModuleNotFoundError:
        raise ImportError(f'{module_name} is required for this module')



from ..base import BaseMapper, TransformElementType

import unicodedata
from typing import Optional, Any, List
from transformers.tokenization_utils_base import PreTrainedTokenizerBase


class TokenizerMapper(BaseMapper):
    def __init__(
            self,
            tokenizer: PreTrainedTokenizerBase,
            input_field: str,
            output_prefix: Optional[str] = None,
            add_special_tokens: Optional[bool] = True,
            max_length: Optional[int] = None,
            is_split_into_words: Optional[bool] = False,
            return_token_type_ids: Optional[bool] = False,
            return_attention_mask: Optional[bool] = True,
            return_overflowing_tokens: Optional[bool] = False,
            return_special_tokens_mask: Optional[bool] = False,
            return_offsets_mapping: Optional[bool] = False,
            return_length: Optional[bool] = False,
            return_word_ids: Optional[bool] = False,
            return_words: Optional[bool] = False,
            **tk_kwargs: Any
    ) -> None:
        super().__init__()

        self.batched = False
        self.tokenizer = tokenizer
        self.prefix = f'{output_prefix}_' if output_prefix else ''

        tk_kwargs = {'add_special_tokens': add_special_tokens,
                     'max_length': max_length,
                     'is_split_into_words': is_split_into_words,
                     **(tk_kwargs or {})}

        input_fields = [input_field]
        output_fields = [f'{self.prefix}input_ids']

        # various options for the tokenizer affect which fields are returned
        tk_kwargs['return_attention_mask'] = return_attention_mask
        if return_attention_mask:
            output_fields.append(f'{self.prefix}attention_mask')

        tk_kwargs['return_token_type_ids'] = return_token_type_ids
        if return_token_type_ids:
            output_fields.append(f'{self.prefix}token_type_ids')

        tk_kwargs['return_overflowing_tokens'] = return_overflowing_tokens
        if return_overflowing_tokens:
            output_fields.append(f'{self.prefix}overflow_to_sample_mapping')

        tk_kwargs['return_special_tokens_mask'] = return_special_tokens_mask
        if return_special_tokens_mask:
            output_fields.append(f'{self.prefix}special_tokens_mask')

        tk_kwargs['return_offsets_mapping'] = return_offsets_mapping
        if return_offsets_mapping:
            output_fields.append(f'{self.prefix}offset_mapping')

        tk_kwargs['return_length'] = return_length
        if return_length:
            output_fields.append(f'{self.prefix}length')

        if 'is_split_into_words' in tk_kwargs and return_word_ids:
            output_fields.append(f'{self.prefix}word_ids')
            if return_words:
                output_fields.append(f'{self.prefix}words')

        self.tokenize_kwargs = tk_kwargs
        self.input_fields = input_fields
        self.output_fields = output_fields

    def transform(self, data: TransformElementType) -> TransformElementType:
        batch_encoding = self.tokenizer(data[self.input_fields[0]], **self.tokenize_kwargs)
        # token_to_word mappings are unfortunately not natively returned by HF.tokenizer
        # so we need to operate separately on the `batch_encoding` object to get this info.
        if f'{self.prefix}word_ids' in self.output_fields:
            word_idss = [
                batch_encoding[sequence_id].word_ids
                for sequence_id in range(len(batch_encoding['input_ids']))
            ]
            if f'{self.prefix}words' in self.output_fields:
                wordss = [
                    [
                        None if word_id is None else data[self.input_fields[0]][word_id]
                        for word_id in word_ids
                    ]
                    for word_ids in word_idss
                ]
            else:
                wordss = None
        else:
            word_idss = None
            wordss = None
            
        # cast to dict
        batch_encoding_as_dict = {f'{self.prefix}{k}': v for k, v in batch_encoding.items()}
        if word_idss is not None:
            batch_encoding_as_dict[f'{self.prefix}word_ids'] = word_idss
            if wordss is not None:
                batch_encoding_as_dict[f'{self.prefix}words'] = wordss
        return dict(batch_encoding_as_dict)


class ValidUnicodeMapper(BaseMapper):
    """Given input_fields of type List[str], replaces invalid Unicode
    characters with something else"""

    def __init__(
            self,
            input_fields: List[str],
            unicode_categories: List[str],
            replace_token: str,
            output_prefix: Optional[str] = None,
    ):
        self.batched = False
        self.input_fields = input_fields
        self.prefix = f'{output_prefix}_' if output_prefix else ''
        self.output_fields = [
            f'{self.prefix}{input_field}' for input_field in self.input_fields
        ]
        self.unicode_categories = unicode_categories
        self.replace_token = replace_token

    def transform(self, data: TransformElementType) -> TransformElementType:
        def _transform(tokens: List[str]) -> List[str]:
            return [
                self.replace_token if all(
                    unicodedata.category(ch) in self.unicode_categories
                    for ch in token
                ) else token
                for token in tokens
            ]
        new_data = {f'{self.prefix}{k}': v if k not in self.input_fields
                    else _transform(v) for k, v in data.items()}
        return new_data



import itertools
import random
from typing import (Any, Iterable, List, Literal, Optional,
                    Sequence, Tuple, TypeVar)
from transformers.tokenization_utils_base import PreTrainedTokenizerBase

from ..base import BaseMapper, TransformElementType


LabelType = TypeVar('LabelType', int, float)


class TokensSequencesPaddingMapper(BaseMapper):
    def __init__(
        self,
        tokenizer: PreTrainedTokenizerBase,
        input_field: str = 'input_ids',
    ) -> None:
        """ Mapper that add BOS/SEP/EOS sequences of tokens.

        Args:
            tokenizer (PreTrainedTokenizerBase): Tokenizer to use for
                looking up special BOS/SEP/EOS tokens.
            input_field (str, optional): The field to add special tokens to.
                Defaults to 'input_ids'.
        """
        super().__init__()

        self.input_fields = [input_field]
        self.output_fields = [input_field]
        self.bos, self.sep, self.eos = self._find_special_token_ids(tokenizer)
        self.batched = False

    @staticmethod
    def _find_special_token_ids(
        tokenizer: PreTrainedTokenizerBase
    ) -> Tuple[Sequence[int], Sequence[int], Sequence[int]]:
        """By default, tokenizers only know how to concatenate 2 fields
        as input; However, for our purposes, we might care about more than
        just 2. This function tries to figure out the best strategy by
        tokenizing two fake sequences and selecting beginning, mid, and
        end sequence(s) tokens."""

        bos, sep, eos = [], [], []

        class FirstFakeSequenceSymbol(int):
            ...

        class SecondFakeSequenceSymbol(int):
            ...

        input_ids = tokenizer.build_inputs_with_special_tokens(
            [FirstFakeSequenceSymbol()], [SecondFakeSequenceSymbol()]
        )

        # anything before the first symbol goes into BOS
        seq_to_append_to = bos
        for token in input_ids:
            if isinstance(token, FirstFakeSequenceSymbol):
                # we found the first fake symbol! switch to
                # mid representation
                seq_to_append_to = sep
            elif isinstance(token, SecondFakeSequenceSymbol):
                # we found the second fake symbol! now we are
                # dealing with EOS
                seq_to_append_to = eos
            else:
                # this is a special token symbol
                seq_to_append_to.append(token)

        return bos, sep, eos

    def transform(self, data: TransformElementType) -> TransformElementType:
        sequences = data[self.input_fields[0]]
        seqs_count = len(sequences)

        padded_sequences = [
            (self.bos if i == 0 else []) +
            seq +
            (self.eos if (i + 1) == seqs_count else self.sep)
            for i, seq in enumerate(sequences)
        ]
        data[self.input_fields[0]] = padded_sequences

        return data


class AttentionMaskSequencePaddingMapper(TokensSequencesPaddingMapper):
    def __init__(self,
                 tokenizer: PreTrainedTokenizerBase,
                 input_field: str = 'attention_mask') -> None:
        """Mapper to add BOS/SEP/EOS tokens to an attention mask sequence.

        Args:
            tokenizer (PreTrainedTokenizerBase): Tokenizer to use for
                looking up size of special BOS/SEP/EOS tokens.
            input_field (str, optional): The field to add special tokens to.
                Defaults to 'attention_mask'.
        """
        super().__init__(tokenizer=tokenizer, input_field=input_field)

        # attention masks are always masked with ones
        self.bos = [1 for _ in self.bos]
        self.sep = [1 for _ in self.sep]
        self.eos = [1 for _ in self.eos]


class TokenTypeIdsSequencePaddingMapper(TokensSequencesPaddingMapper):
    def __init__(self,
                 tokenizer: PreTrainedTokenizerBase,
                 input_field: str = 'token_type_ids') -> None:
        """Mapper to add BOS/SEP/EOS tokens to a token type ids sequence.

        Args:
            tokenizer (PreTrainedTokenizerBase): Tokenizer to use for
                looking up size of special BOS/SEP/EOS tokens.
            input_field (str, optional): The field to add special tokens to.
                Defaults to 'token_type_ids'.
        """
        super().__init__(tokenizer=tokenizer, input_field=input_field)

    def transform(self, data: TransformElementType) -> TransformElementType:
        sequences = data[self.input_fields[0]]
        seqs_count = len(sequences)
        padded_sequences = [
            # a sequence start with BOS tags or SEP tags
            ([i for _ in self.bos] if i == 0 else [i for _ in self.sep]) +
            seq +
            # a sequence ends with EOS tags or nothing if it is not
            # the last sequence
            ([i for _ in self.eos] if (i + 1) == seqs_count else [])
            for i, seq in enumerate(sequences)
        ]
        data[self.input_fields[0]] = padded_sequences

        return data


class MakeAttentionMaskMapper(BaseMapper):
    def __init__(self,
                 input_field: str = 'input_ids',
                 output_field: str = 'attention_mask') -> None:
        """Mapper to create attention masks from input ids.

        Args:
            input_field (str, optional): The field to determine the
                shape of the attention mask. Defaults to 'input_ids'.
            output_field (str, optional): The name of the field containing
                the attention mask. Defaults to 'attention_mask'.
        """
        super().__init__()
        self.input_fields = [input_field]
        self.output_fields = [output_field]

    def transform(self, data: TransformElementType) -> TransformElementType:
        sequences = data[self.input_fields[0]]
        attention_masks = [[1 for _ in range(len(seq))] for seq in sequences]
        data[self.output_fields[0]] = attention_masks
        return data


class LabelsMaskerMapper(BaseMapper):
    def __init__(self,
                 labels_field: str = 'labels',
                 strategy: Literal['all', 'one', 'sample'] = 'all',
                 sample_prob: Optional[float] = None,
                 label_mask_id: LabelType = -100) -> None:
        """ Given a sequence of labels, this mapper will mask some of them.
        Useful when wanting to create more samples by masking a subset of
        labels, or when running evaluations and want to predict one label at
        the time.

        Args:
            labels_field (str, optional): The field containing the labels.
                Defaults to 'labels'.
            strategy (Literal['all', 'one', 'sample'], optional): The strategy
                to use for masking. If 'all', no values are masked. If 'one',
                we generate m separate samples for each label, where for each
                sample, we mask one label. If 'sample', we generate sample
                a subset of labels with a probability of `sample_prob`.
                Defaults to 'all'.
            sample_prob (float, optional): The probability of sampling when
                `strategy` is 'sample'. Defaults to None. If strategy is
                'sample' and sample_prob is None, an error is raised.
            label_mask_id (LabelType, optional): The value to use for a masked
                label. Defaults to -100.

        """
        super().__init__()
        self.input_fields = [labels_field]
        self.output_fields = [labels_field]
        self.batched = True

        if strategy not in ['all', 'one', 'sample']:
            raise ValueError(f'Unknown strategy {strategy}')
        elif strategy == 'random' and sample_prob is None:
            raise ValueError('no `sample_prob` provided for `random` strategy')
        elif strategy == 'one' and sample_prob is not None:
            raise ValueError('Do not provide `sample_prob` for `one` strategy')

        self.strategy: str = strategy
        self.sample_prob: float = sample_prob or 0.0
        self.label_mask_id: LabelType = label_mask_id

    def transform(
        self,
        data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:

        if self.strategy == 'all':
            # there's no masking to do if the strategy is all!
            # `all` is provided for convenience, but it's not really a
            # transformation, just an identity function
            yield from data

        labels_field_name, *_ = self.input_fields

        for sample in data:
            labels = sample[labels_field_name]

            if self.strategy == 'one':
                # make sequences of labels where only one label is
                # not masked for each sequence; the goal is to transform
                # n sequences with m active labels into n * m sequences
                # with only one active label.
                for i, _ in enumerate(labels):

                    # new labels sequence here
                    new_labels = [l if i == j else self.label_mask_id
                                  for j, l in enumerate(labels)]

                    new_sample = {**sample,
                                  **{labels_field_name: new_labels}}
                    yield new_sample

                    # # add the new sample you just made to the batch
                    # new_data[labels_field_name].append(new_labels)
                    # for f in other_fields_name:
                    #     new_data[f].append(data[f][sample_pos])

            if self.strategy == 'sample':
                expected_slice_size = max(
                    int(len(labels) * self.sample_prob), 1
                )
                labels_pos = list(range(len(labels)))
                random.shuffle(labels_pos)

                for active_pos in (
                    labels_pos[i:i + expected_slice_size]
                    for i in range(0, len(labels_pos), expected_slice_size)
                ):
                    # new labels sequence here
                    new_labels = [l if i in active_pos else self.label_mask_id
                                  for i, l in enumerate(labels)]

                    # add the new sample you just made to the batch,
                    # plus all extra fields
                    new_sample = {**sample,
                                  **{labels_field_name: new_labels}}
                    yield new_sample


class MultiSequenceStriderMapper(BaseMapper):
    def __init__(self,
                 max_stride_count: int,
                 length_reference_field: str,
                 max_length: Optional[int] = None,
                 extra_length_per_seq: Optional[int] = None,
                 tokenizer: Optional[PreTrainedTokenizerBase] = None,
                 max_step: Optional[int] = None) -> None:
        """Mapper to create multiple subset sequences from a single sequence
        of sequences.

        The multiple sequences are created by sliding a window of size
        `max_stride_count` over the original sequence or `max_length` if
        the total number of accumulated tokens exceeds it (whichever comes
        first).

        Args:
            max_stride_count (int): The maximum number of sequences to include
                in each subset sequence of sequences.
            length_reference_field (str): The field to use to determine the
                rolling length of a subset sequence.
            max_length (int, optional): The maximum length of units in
                the subset sequence. Defaults to None (i.e., not used).
            extra_length_per_seq (int, optional): Optional field in case
                you we expect each sequence to be extended by another mapper
                after MultiSequenceStriderMapper. For example, if you are
                expecting to add special tokens for BOS/EOS/SEP,
                extra_length_per_seq could be 2 if BOS, EOS, and SEP are
                length 1.
            tokenizer (PreTrainedTokenizerBase, optional): A HuggingFace
                tokenizer to use to determine the length of BOS/EOS/SEP
                tokens. If not provided, extra_length_per_seq is used.
            max_step (int, optional): Not used at the moment.

        """
        super().__init__()

        self.input_fields = self.output_fields = [length_reference_field]
        self.batched = True
        self.max_stride_count = max_stride_count
        self.max_length = max_length or float('inf')

        if extra_length_per_seq is None:
            if tokenizer is not None:
                # if a tokenizer is provided, we use the extra
                # length we need for each sequence is to account for
                # padding of sequences. Because we don't know if a sequence
                # will be at the beginning, middle, or end of the batch,
                # we take the max length between the various separators
                bos, sep, eos = TokensSequencesPaddingMapper.\
                    _find_special_token_ids(tokenizer)
                extra_length_per_seq = 2 * max(len(bos), len(sep), len(eos))
            else:
                # if extra length is not provided, we simply assume that
                # there is not need to account for extra length that is
                # needed for special token separation (e.g. BOS/EOS symbols)
                extra_length_per_seq = 0
        self.extra_length_per_seq = extra_length_per_seq

        if max_step is not None:
            # TODO: implement max_step to support overlapping strides
            raise NotImplementedError("max_step is not supported yet")

    def transform(
        self,
        data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:

        ref_field_name, *_ = self.input_fields

        for sample in data:
            seq_pos_start = 0
            cumulative_stride_length = 0

            for seq_pos_end in range(len(sample[ref_field_name])):
                current_seq_length = (
                    len(sample[ref_field_name][seq_pos_end]) +
                    self.extra_length_per_seq
                )

                if current_seq_length > self.max_length:
                    raise ValueError(
                        "Current sequence is longer than the maximum stride"
                        f"length ({current_seq_length} > {self.max_length})"
                    )

                stride_too_long = (
                    (cumulative_stride_length + current_seq_length)
                    > self.max_length
                )
                stride_has_too_many_seqs = (
                    (seq_pos_end - seq_pos_start)
                    >= self.max_stride_count
                )

                if stride_too_long or stride_has_too_many_seqs:
                    yield {k: v[seq_pos_start:seq_pos_end]
                           for k, v in sample.items()}

                    cumulative_stride_length = 0
                    seq_pos_start = seq_pos_end

                # now that the current sequence is included in the next
                # stride, we add its length to the cumulative stride length
                cumulative_stride_length += current_seq_length

            # yield the last sequence
            out = {k: v[seq_pos_start:] for k, v in sample.items()}

            if len(out[ref_field_name]) < 1:
                import ipdb
                ipdb.set_trace()
            yield out


class SingleValueToSequenceMapper(BaseMapper):
    def __init__(self,
                 single_value_field: str,
                 like_field: str = 'input_ids',
                 strategy: Literal['first', 'last', 'all'] = 'first',
                 padding_id: LabelType = -100) -> None:
        """Mapper to create a sequence of values from single value.
        Useful when casting a sequence classification task to a sequence
        tagging task, e.g. making a prediction for a sequence of sentences
        by concatenating the sentences, and then predicting on the BOS/SEP
        tokens.

        Args:
            single_value_field: name of the field containing the single
                label value.
            like_field: name of the field whose shape will be used to repeat
                the single value to create the sequence. Default is 'input_ids'.
            strategy: strategy to use to create the sequence.
                - If 'first', the single value will be the first element of the
                    new sequence, and all other positions will be filled with
                    the padding_id.
                - If 'last', the single value will be the last element of the
                    new sequence, and all other positions will be filled with
                    the padding_id.
                - If 'all', the single value will be repeated for each position
                    of the new sequence; the padding_id will be ignored.
            padding_id: id to use for the padding token. Default is -100.
        """
        super().__init__()

        self.input_fields = [single_value_field, like_field]
        self.output_fields = [single_value_field]
        self.strategy = strategy
        self.padding_id = padding_id
        self.batched = False

    def _make_sequence_from_value(
        self,
        value: LabelType,
        like_seq: Sequence[Any]
    ) -> Sequence[LabelType]:

        if self.strategy == 'first':
            return ([value] +       # type: ignore
                    [self.padding_id for _ in range(len(like_seq) - 1)])
        elif self.strategy == 'last':
            return ([self.padding_id for _ in range(len(like_seq) - 1)] +
                    [value])        # type: ignore
        elif self.strategy == 'all':
            return [value for _ in like_seq]
        else:
            raise ValueError(f'Strategy {self.strategy} is not supported')

    def transform(self, data: TransformElementType) -> TransformElementType:
        labels_field_name, like_field_name, *_ = self.input_fields

        data[labels_field_name] = [
            self._make_sequence_from_value(
                value=label, like_seq=data[like_field_name][i]
            ) for i, label in enumerate(data[labels_field_name])
        ]
        return data


class SequencesConcatenateMapper(BaseMapper):
    def __init__(self, concat_fields: Optional[List[str]] = None):
        super().__init__()

        self.concat_fields = (
            set(concat_fields) if concat_fields is not None else None
        )
        self.input_fields = concat_fields or []
        self.output_fields = concat_fields or []
        self.batched = False

    def _to_concat(self, field_name: str) -> bool:
        return self.concat_fields is None or field_name in self.concat_fields

    def transform(self, data: TransformElementType) -> TransformElementType:
        return {key: (list(itertools.chain.from_iterable(value))
                      if self._to_concat(key) else value)
                for key, value in data.items()}



from .dataset import BaseDataset
from .mapper import BaseMapper
from .types import (
    Features, FeatureType, TransformBatchType, TransformElementType
)

__all__ = [
    'BaseDataset',
    'BaseMapper',
    'Features',
    'FeatureType',
    'TransformBatchType',
    'TransformElementType',
]



from typing import Callable, Optional, Protocol, TypeVar

from .types import FeatureType, Features


D = TypeVar('D', bound='BaseDataset')


class BaseDataset(Protocol):
    """A protocol for datasets. In general, a dataset is a collection of samples
    that can be iterated over. Datasets support a map operation that can either
    transform a sample into another sample (1-to-1 mapping) or generate
    multiple samples from a single sample (1-to-many mapping).

    Dataset APIs are modeled after HuggingFace ArrowDataset APIs."""


    @property
    def features(self) -> Features:
        ...

    def map(self: D,
            function: Optional[Callable] = None,
            *,
            batched: bool = False,
            batch_size: Optional[int] = 1000) -> D:
        ...

    def cast_column(
        self: D,
        column: str,
        feature: FeatureType,
        *_
    ) -> D:
        ...



from ...base import BaseMapper, TransformElementType


class OneVsOtherAnnotatorMapper(BaseMapper):
    def __init__(self,
                 input_field: str = 'annotations',
                 label_field: str = 'labels',
                 preds_field: str = 'preds',
                 position: int = 0) -> None:
        super().__init__()
        self.input_fields = [input_field]
        self.output_fields = [label_field, preds_field]
        self.batched = False
        self.position = position

    def transform(self, data: TransformElementType) -> TransformElementType:
        input_field_name, *_ = self.input_fields
        label_field_name, preds_field_name, *_ = self.output_fields

        if len(data[input_field_name]) < 2:
            raise ValueError(f'Expected at least 2 annotations, '
                             f'got {len(data[input_field_name])}')

        other_labels = [v for i, v in enumerate(data[input_field_name])
                        if i != self.position]

        return {preds_field_name: data[input_field_name][self.position],
                label_field_name: sum(other_labels) / len(other_labels)}




from typing import Any, Optional, Sequence, Callable

from ..base import BaseMapper, TransformElementType, BaseDataset


class ChangeFieldsMapper(BaseMapper):
    def __init__(self,
                 keep_fields:  Optional[Sequence[str]] = None,
                 drop_fields: Optional[Sequence[str]] = None):
        """Mapper that removes some of the fields in a dataset.
        Either `keep_fields` or `drop_fields` must be specified, but not both.

        Args:
            keep_fields (Sequence[str]): Fields to keep, all other fields
                are dropped. Defaults to [].
            drop_fields (Sequence[str]): Fields to drop, all other fields
                are kept. Defaults to [].
        """

        # xor between keep_fields and remove_fields
        if (keep_fields is not None and drop_fields is not None) or \
                (keep_fields is None and drop_fields is None):
            raise ValueError('Must specify `keep_fields` or `drop_fields`')

        self.input_fields = drop_fields or []
        self.output_fields = keep_fields or []

        self.batched = False

    def map(self,
            dataset: BaseDataset,
            **map_kwargs: Any) -> BaseDataset:
        map_kwargs = {'remove_columns': list(dataset.features.keys()),
                      **map_kwargs}
        return super().map(dataset, **map_kwargs)

    def transform(self, data: TransformElementType) -> TransformElementType:
        if self.input_fields:
            new_data = {k: v for k, v in data.items()
                        if k not in self.input_fields}

        elif self.output_fields:
            new_data = {k: data[k] for k in self.output_fields}

        else:
            raise ValueError('Must specify `keep_fields` or `drop_fields`')

        return new_data


class MakeFieldMapper(BaseMapper):
    def __init__(
        self,
        field_name: str,
        value: Optional[Any] = None,
        shape_like: Optional[str] = None,
        value_fn: Optional[Callable[[TransformElementType], Any]] = None
    ):
        """Mapper that adds a new field to a dataset.
        Either `value` or `value_fn` must be specified, but not both.
        `value` is a constant value to assign to the new field, while
        `value_fn` is a function that takes the full sample as input and
        returns a value to assign to the new field.

        Args:
            field_name (str): Name of the new field.
            value (Optional[Any], optional): Value to assign to the new field.
            shape_like (Optional[str], optional): If a fixed value is provided,
                this existing field that will be used to determine the shape of
                the new field. Defaults to None.
            value_fn (Optional[Callable[[TransformElementType], Any]], optional):
                Function to call to assign a value to the new field.
                Defaults to None.
        """

        self.input_fields = []
        self.output_fields = [field_name]
        self.batched = False

        if value_fn is None:
            if value is None:
                raise ValueError('Must specify `value` or `value_fn`')

            def _value_fn(data: TransformElementType) -> Any:
                if shape_like is not None:
                    return [value for _ in data[shape_like]]
                else:
                    return value

            # doing the two assignment separately to avoid linter
            # error about redefinition of `value_fn` via def
            value_fn = _value_fn

        self.value_fn = value_fn

    def transform(self, data: TransformElementType) -> TransformElementType:
        data[self.output_fields[0]] = self.value_fn(data)
        return data



from typing import Any, Callable, Iterable, Optional, Sequence, Type


from ..base import (BaseDataset, BaseMapper, Features, FeatureType,
                    TransformElementType)
from ..mappers import multiseq, shape, tokenize, fields
from ..mappers.contrib import sse


__all__ = [
    'Dataset',
    'TokensSequencesPaddingMapper',
    'AttentionMaskSequencePaddingMapper',
    'TokenTypeIdsSequencePaddingMapper',
    'MakeAttentionMaskMapper',
    'LabelsMaskerMapper',
    'MultiSequenceStriderMapper',
    'SingleValueToSequenceMapper',
    'SequencesConcatenateMapper',
    'FlattenMapper',
    'BinarizerMapper',
    'TokenizerMapper',
    'OneVsOtherAnnotatorMapper',
    'ChangeFieldsMapper',
    'MakeFieldMapper',
    'ValidUnicodeMapper'
]


class Dataset(list, BaseDataset):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def features(self) -> Features:
        if len(self):
            return {k: type(v) for k, v in self[0].items()}
        else:
            raise ValueError('Dataset is empty; cant get features')

    def keys(self) -> Sequence[str]:
        return [k for k in self.features.keys()]

    def map(
        self: 'Dataset',
        function: Optional[Callable] = None,
        batched: bool = False,
        *args,
        **kwargs
    ) -> 'Dataset':
        if function is None:
            return self

        processed = Dataset()
        for sample in self:
            if batched:
                processed.extend(function([sample]))
            else:
                processed.append(function(sample))

        return processed

    def cast_column(
        self: 'Dataset',
        column: str,
        feature: FeatureType,
        *_
    ) -> 'Dataset':

        def cast_fn(sample: TransformElementType) -> TransformElementType:
            return {k: (feature(v) if k == column else v)
                    for k, v in sample.items()}

        return self.map(cast_fn)


class _SimpleInterfaceMixInMapper(BaseMapper):
    def map(self, dataset: Dataset, **map_kwargs: Any) -> Dataset:
        return super().map(dataset, **map_kwargs)

    @classmethod
    def chain(cls: Type['_SimpleInterfaceMixInMapper'],
              dataset: Dataset,
              mappers: Sequence['_SimpleInterfaceMixInMapper'],
              **map_kwargs: Any) -> Dataset:
        return super().chain(dataset, mappers, **map_kwargs)

    def batch_transform(
        self: '_SimpleInterfaceMixInMapper',
        data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:
        # Simple mappers are called with slightly different data from
        # the base class, so we need to override this method.
        yield from super().transform(data)


class TokensSequencesPaddingMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.TokensSequencesPaddingMapper
):
    ...


class AttentionMaskSequencePaddingMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.AttentionMaskSequencePaddingMapper
):
    ...


class TokenTypeIdsSequencePaddingMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.TokenTypeIdsSequencePaddingMapper
):
    ...


class MakeAttentionMaskMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.MakeAttentionMaskMapper
):
    ...


class LabelsMaskerMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.LabelsMaskerMapper
):
    ...


class MultiSequenceStriderMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.MultiSequenceStriderMapper
):
    ...


class SingleValueToSequenceMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.SingleValueToSequenceMapper
):
    ...


class SequencesConcatenateMapper(
    _SimpleInterfaceMixInMapper,
    multiseq.SequencesConcatenateMapper
):
    ...


class FlattenMapper(
    _SimpleInterfaceMixInMapper,
    shape.FlattenMapper
):
    ...


class BinarizerMapper(
    _SimpleInterfaceMixInMapper,
    shape.BinarizerMapper
):
    ...


class TokenizerMapper(
    _SimpleInterfaceMixInMapper,
    tokenize.TokenizerMapper
):
    ...


class OneVsOtherAnnotatorMapper(
    _SimpleInterfaceMixInMapper,
    sse.OneVsOtherAnnotatorMapper
):
    ...


class ChangeFieldsMapper(
    _SimpleInterfaceMixInMapper,
    fields.ChangeFieldsMapper
):
    ...


class MakeFieldMapper(
    _SimpleInterfaceMixInMapper,
    fields.MakeFieldMapper
):
    ...

class ValidUnicodeMapper(
    _SimpleInterfaceMixInMapper,
    tokenize.ValidUnicodeMapper
):
    ...



from abc import ABC, abstractmethod
from typing import \
    Any, Dict, Iterable, Optional, Sequence, Tuple, Type, overload

from .dataset import D
from .types import \
    TransformBatchType, TransformElementType, Features, FeatureType


class BaseMapper(ABC):
    """An abstract implementation of a Mapper; must be subclassed.

    A mapper is an object that transforms samples in a dataset; it can either
    produce one transformed sample per input sample (1-to-1 mapping,
    self.batched = False) or produce n samples per input sample (1-to-n mapping,
    self.batched = True).

    When creating a mapper from this base class, one need to assign three
    attributes:
    - self.input_fields: a list of the names of the input fields that are
        required to be present in the dataset for the mapper to work.
    - self.output_fields: a list of the names of the output fields that are
        produced by the mapper.
    - self.batched: a boolean indicating whether the mapper produces one
        transformed sample per input sample (False) or n samples per input
        sample (True).

    Beside the above attributes, the mapper also needs to implement the
    transform method, which either receives a single sample (if
    self.batched = False) or a list of samples (if self.batched = True).

    Optionally, it can also implement the cast_columns method, which returns
    a dictionary of new types for some (or all) of the input/output columns.
    """

    __batched__: Optional[bool] = None
    __input_fields__: Optional[Sequence[str]] = None
    __output_fields__: Optional[Sequence[str]] = None

    @property
    def batched(self) -> bool:
        if self.__batched__ is None:
            raise NotImplementedError('Mapper subclass must set batched')
        return self.__batched__

    @batched.setter
    def batched(self, value: bool):
        self.__batched__ = bool(value)

    @property
    def input_fields(self) -> Sequence[str]:
        if self.__input_fields__ is None:
            raise NotImplementedError('Mapper subclass must set input_fields')
        return tuple(str(e) for e in self.__input_fields__)

    @input_fields.setter
    def input_fields(self, input_fields: Sequence[str]):
        self.__input_fields__ = tuple(str(e) for e in input_fields)

    @property
    def output_fields(self) -> Sequence[str]:
        if self.__output_fields__ is None:
            raise NotImplementedError('Mapper subclass must set output_fields')
        return tuple(str(e) for e in self.__output_fields__)

    @output_fields.setter
    def output_fields(self, output_fields: Sequence[str]):
        self.__output_fields__ = tuple(str(e) for e in output_fields)

    def verify_output(self, data: Dict[str, Any]):
        for field in self.output_fields:
            if field not in data:
                raise ValueError(f'Field {field} not found in returned data')

    @classmethod
    def chain(cls: Type['BaseMapper'],
              dataset: D,
              mappers: Sequence['BaseMapper'],
              **map_kwargs: Any) -> D:
        for mapper in mappers:
            dataset = mapper.map(dataset, **map_kwargs)
        return dataset

    def map(self: 'BaseMapper',
            dataset: D,
            **map_kwargs: Any) -> D:
        for field in self.input_fields:
            if field not in dataset.features:
                msg = f'Field {field} not found in input dataset {dataset}'
                raise ValueError(msg)

        map_kwargs = {'batched': self.batched, **map_kwargs}

        if self.batched:
            dataset = dataset.map(function=self.batch_transform, **map_kwargs)
        else:
            dataset = dataset.map(function=self.transform, **map_kwargs)

        # We optionally cast some (or all!) of the output columns
        # to a new format. By default, no casting occurs.
        to_cast_features = self.cast_columns(dataset.features)
        for feat_name, feat_type in to_cast_features.items():
            dataset = dataset.cast_column(feat_name, feat_type)

        for field in self.output_fields:
            if field not in dataset.features:
                raise ValueError(f'Field {field} not found in returned data')

        return dataset

    def batch_transform(self: 'BaseMapper',
                        data: TransformBatchType) -> TransformBatchType:
        keys = [k for k in data.keys()]

        def _index_fn(t: Tuple[str, Any]) -> int:
            k, _ = t
            return keys.index(k)

        to_transform_iterable = (
            dict(zip(keys, sample)) for sample in
            zip(*(v for _, v in sorted(data.items(), key=_index_fn)))
        )
        transformed_batch = {}
        for transformed_sample in self.transform(to_transform_iterable):
            for k, v in transformed_sample.items():
                transformed_batch.setdefault(k, []).append(v)

        return transformed_batch

    @abstractmethod
    @overload
    def transform(
        self: 'BaseMapper',
        data: TransformElementType
    ) -> TransformElementType:
        ...

    @abstractmethod
    @overload
    def transform(
        self: 'BaseMapper',
        data: Iterable[TransformElementType]
    ) -> Iterable[TransformElementType]:
        ...

    @abstractmethod
    def transform(self, data):
        raise NotImplementedError('Mapper subclass must implement transform')

    def cast_columns(self, features: Features) -> Dict[str, FeatureType]:
        return {}



from itertools import chain
from typing import Dict

from ..base import BaseMapper, TransformElementType, Features, FeatureType


class FlattenMapper(BaseMapper):
    def __init__(self, field: str) -> None:
        super().__init__()
        self.input_fields = [field]
        self.output_fields = [field]
        self.batched = False

    def transform(self, data: TransformElementType) -> TransformElementType:
        field_name, *_ = self.input_fields

        flattened_field = data[field_name]

        if len(flattened_field) > 0:
            while isinstance(flattened_field[0], list):
                flattened_field = list(chain.from_iterable(flattened_field))

        return {field_name: flattened_field}


class BinarizerMapper(BaseMapper):
    __value_type__: type = str
    __sequence_type__: type = list

    def __init__(self, field: str, threshold: float) -> None:
        super().__init__()
        self.input_fields = [field]
        self.output_fields = [field]
        self.threshold = threshold
        self.batched = False

    def cast_columns(self, features: Features) -> Dict[str, FeatureType]:
        field_name, *_ = self.input_fields
        if isinstance(features[field_name], self.__sequence_type__):
            new_field = self.__sequence_type__(self.__value_type__('int64'))
        else:
            new_field = self.__value_type__('int64')

        return {field_name: new_field}

    def transform(self, data: TransformElementType) -> TransformElementType:
        field_name, *_ = self.input_fields

        if isinstance(data[field_name], list):
            binarized_field = [1 if v > self.threshold else 0
                               for v in data[field_name]]
        else:
            binarized_field = 1 if data[field_name] > self.threshold else 0

        return {field_name: binarized_field}



from ..utils import requires

requires("datasets")

from typing import Any, Sequence, Type, TypeVar, Dict

from datasets.arrow_dataset import Dataset
from datasets.iterable_dataset import IterableDataset
from datasets.features.features import Features, FeatureType


from ..base import BaseMapper
from ..mappers import multiseq, shape, tokenize, fields
from ..mappers.contrib import sse

from datasets import features


HfDatasetType = TypeVar("HfDatasetType", Dataset, IterableDataset)


__all__ = [
    'TokensSequencesPaddingMapper',
    'AttentionMaskSequencePaddingMapper',
    'TokenTypeIdsSequencePaddingMapper',
    'MakeAttentionMaskMapper',
    'LabelsMaskerMapper',
    'MultiSequenceStriderMapper',
    'SingleValueToSequenceMapper',
    'SequencesConcatenateMapper',
    'FlattenMapper',
    'BinarizerMapper',
    'TokenizerMapper',
    'OneVsOtherAnnotatorMapper',
    'ChangeFieldsMapper',
    'ValidUnicodeMapper'
]


class _HuggingFaceInterfaceMixInMapper(BaseMapper):

    def map(self, dataset: Dataset, **map_kwargs: Any) -> Dataset:
        return super().map(dataset, **map_kwargs)

    @classmethod
    def chain(cls: Type['_HuggingFaceInterfaceMixInMapper'],
              dataset: HfDatasetType,
              mappers: Sequence['_HuggingFaceInterfaceMixInMapper'],
              **map_kwargs: Any) -> HfDatasetType:
        ...

    def cast_columns(self, features: Features) -> Dict[str, FeatureType]:
        return super().cast_columns(features)


class TokensSequencesPaddingMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.TokensSequencesPaddingMapper
):
    ...


class AttentionMaskSequencePaddingMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.AttentionMaskSequencePaddingMapper
):
    ...


class TokenTypeIdsSequencePaddingMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.TokenTypeIdsSequencePaddingMapper
):
    ...


class MakeAttentionMaskMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.MakeAttentionMaskMapper
):
    ...


class LabelsMaskerMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.LabelsMaskerMapper
):
    ...


class MultiSequenceStriderMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.MultiSequenceStriderMapper
):
    ...


class SingleValueToSequenceMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.SingleValueToSequenceMapper
):
    ...


class SequencesConcatenateMapper(
    _HuggingFaceInterfaceMixInMapper,
    multiseq.SequencesConcatenateMapper
):
    ...


class FlattenMapper(
    _HuggingFaceInterfaceMixInMapper,
    shape.FlattenMapper
):
    ...


class BinarizerMapper(
    _HuggingFaceInterfaceMixInMapper,
    shape.BinarizerMapper
):
    __value_type__: type = features.Value
    __sequence_type__: type = features.Sequence


class TokenizerMapper(
    _HuggingFaceInterfaceMixInMapper,
    tokenize.TokenizerMapper
):
    ...


class OneVsOtherAnnotatorMapper(
    _HuggingFaceInterfaceMixInMapper,
    sse.OneVsOtherAnnotatorMapper
):
    ...


class ChangeFieldsMapper(
    _HuggingFaceInterfaceMixInMapper,
    fields.ChangeFieldsMapper
):
    ...


class ValidUnicodeMapper(
    _HuggingFaceInterfaceMixInMapper,
    tokenize.ValidUnicodeMapper
):
    ...
